@@ -1,12 +1,12 @@
-"""Python utilities to simplify calls to some parts of the Data Manager API.
+#!/usr/bin/env python
+"""Python utilities to simplify calls to some parts of the Data Manager API to
+interact with **Projects**, **Instances** (**Jobs**) and **Files**.
 
-For API methods where a user can expect material from a successful call
-the original response payload can be found in the DmApiRv 'msg' property,
-rendered as a Python dictionary.
-
-The URL to the DM API URL is picked up from the environment variable
-'SQUONK_API_URL'. If this isn't set the user can set it programmatically
-using the 'DmApi.set_api_url()' method.
+.. note::
+    The URL to the DM API is picked automatically up from the environment variable
+    ``SQUONK_API_URL``, expected to be of the form **https://example.com/data-manager-api**.
+    If the variable isn't set the user must set it programmatically
+    using :py:meth:`DmApi.set_api_url()`.
 """
 from collections import namedtuple
 from datetime import datetime
@@ -22,11 +22,18 @@ from authlib.jose import jwt
 from wrapt import synchronized
 import requests
 
-# The return value from our public methods.
-# 'success' (a boolean) is True if the call was successful, False otherwise.
-# The 'msg' (an optional dictionary) will provide a potentially useful
-# error message or API-specific response content on success.
 DmApiRv: namedtuple = namedtuple('DmApiRv', 'success msg')
+"""The return value from most of the the DmApi class public methods.
+
+:param success: True if the call was successful, False otherwise.
+:param msg: API request response content
+"""
+
+TEST_PRODUCT_ID: str = 'product-11111111-1111-1111-1111-111111111111'
+"""A test AS Product ID, This ID does not actually exist but is accepted
+as valid by the Data Manager for Administrative users and used for
+testing purposes.
+"""
 
 # The Job instance Application ID - a 'well known' identity.
 _DM_JOB_APPLICATION_ID: str = 'datamanagerjobs.squonk.it'
@@ -42,7 +49,10 @@ _LOGGER: logging.Logger = logging.getLogger(__name__)
 
 
 class DmApi:
-    """Simplified API access methods.
+    """The DmAPI class provides high-level, simplified access to the DM API.
+    You can use the request module directly for finer control. This module
+    provides a wrapper around the handling of the request, returning a simplified
+    namedtuple response value ``DmApiRv``
     """
 
     # The default DM API is extracted from the environment,
@@ -84,7 +94,7 @@ class DmApi:
 
         if not DmApi._dm_api_url:
             return DmApiRv(success=False,
-                           msg={'msg': 'No API URL defined'}), None
+                           msg={'error': 'No API URL defined'}), None
 
         url: str = DmApi._dm_api_url + endpoint
 
@@ -113,7 +123,7 @@ class DmApi:
             _LOGGER.exception('Request failed')
         if resp is None or resp.status_code not in expected_codes:
             return DmApiRv(success=False,
-                           msg={'msg': f'{error_message} (resp={resp})'}),\
+                           msg={'error': f'{error_message} (resp={resp})'}),\
                    resp
 
         # Try and decode the response,
@@ -187,8 +197,11 @@ class DmApi:
     @classmethod
     @synchronized
     def set_api_url(cls, url: str, verify_ssl_cert: bool = True) -> None:
-        """Replaces the class API URL value.
-        Typically https://example.com/data-manager-api
+        """Replaces the API URL value, which is otherwise set using
+        the ``SQUONK_API_URL`` environment variable.
+
+        :param url: The API endpoint, typically **https://example.com/data-manager-api**
+        :param verify_ssl_cert: Use False to avoid SSL verification in request calls
         """
         assert url
         DmApi._dm_api_url = url
@@ -201,7 +214,7 @@ class DmApi:
     @classmethod
     @synchronized
     def get_api_url(cls) -> Tuple[str, bool]:
-        """Return the URL and whether validating the SSL layer.
+        """Return the API URL and whether validating the SSL layer.
         """
         return DmApi._dm_api_url, DmApi._verify_ssl_cert
 
@@ -217,13 +230,24 @@ class DmApi:
                          timeout_s: int = 4)\
             -> Optional[str]:
         """Gets a DM API access token from the given Keycloak server, realm
-        and client ID. The keycloak URL is typically 'https://example.com/auth'.
+        and client ID.
+
         If keycloak fails to yield a token None is returned, with messages
         written to the log.
 
-        The caller can (is encouraged to) provide a prior token. The token
-        logic then only calls keycloak to obtain a new token if the current
+        The caller can (is encouraged to) provide a prior token in oprder to
+        reduce token requests on the server. When a ``prior_token`` is provided
+        the code only calls keycloak to obtain a new token if the current
         one looks like it will expire (in less than 60 seconds).
+
+        :param keycloak_url: The keycloak server URL, typically **https://example.com/auth**
+        :param keycloak_realm: The keycloak realm
+        :param keycloak_client_id: The keycloak DM-API client ID
+        :param username: A valid username
+        :param password: A valid password
+        :param prior_token: An optional prior token. If supplied it will be used
+            unless it is about to expire
+        :param timeout_s: The underlying request timeout
         """
         assert keycloak_url
         assert keycloak_realm
@@ -288,9 +312,11 @@ class DmApi:
     @synchronized
     def ping(cls, access_token: str, timeout_s: int = 4)\
             -> DmApiRv:
-        """Calls the DM API to ensure the server is responding.
-        Here we do something relatively simple, and the best endpoint
-        to call (in DM 0.7) is '/account-server/namespace'.
+        """A handy API method that calls the DM API to ensure the server is
+        responding.
+
+        :param access_token: A valid DM API access token
+        :param timeout_s: The underlying request timeout
         """
         assert access_token
 
@@ -303,7 +329,10 @@ class DmApi:
     @synchronized
     def get_version(cls, access_token: str, timeout_s: int = 4)\
             -> DmApiRv:
-        """Calls the DM API to get the underlying service version.
+        """Returns the DM-API service version.
+
+        :param access_token: A valid DM API access token
+        :param timeout_s: The underlying request timeout
         """
         assert access_token
 
@@ -317,22 +346,26 @@ class DmApi:
     def create_project(cls,
                        access_token: str,
                        project_name: str,
-                       as_tier_product_id: str = 'product-11111111-1111-1111-1111-111111111111',
-                       as_organisation_id: str = 'org-11111111-1111-1111-1111-111111111111',
-                       as_unit_id: str = 'unit-11111111-1111-1111-1111-111111111111',
+                       as_tier_product_id: str,
                        timeout_s: int = 4)\
             -> DmApiRv:
-        """Creates a project using an organisation, unit and product.
+        """Creates a Project, which requires a name and a Product ID obtained from
+        the Account Server.
+
+        :param access_token: A valid DM API access token.
+        :param project_name: A unique name.
+        :param as_tier_product_id: If no account server is
+            attached any suitable value can be used. If you are an admin user
+            you can also use the reserved value of
+            ``product-11111111-1111-1111-1111-111111111111``
+            which is automatically accepted.
+        :param timeout_s: The API request timeout
         """
         assert access_token
         assert project_name
         assert as_tier_product_id
-        assert as_organisation_id
-        assert as_unit_id
 
-        data: Dict[str, Any] = {'unit_id': as_unit_id,
-                                'organisation_id': as_organisation_id,
-                                'tier_product_id': as_tier_product_id,
+        data: Dict[str, Any] = {'tier_product_id': as_tier_product_id,
                                 'name': project_name}
         return DmApi._request('POST', '/project',
                               access_token=access_token,
@@ -349,6 +382,10 @@ class DmApi:
                        timeout_s: int = 4) \
             -> DmApiRv:
         """Deletes a project.
+
+        :param access_token: A valid DM API access token
+        :param project_id: The DM-API project id to delete
+        :param timeout_s: The underlying request timeout
         """
         assert access_token
         assert project_id
@@ -369,12 +406,21 @@ class DmApi:
                                     timeout_per_file_s: int = 120)\
             -> DmApiRv:
         """Puts a file, or list of files, into a DM Project
-        using an optional path. The files can include relative or absolute
-        paths but are written to the same path in the project.
+        using an optional path.
 
-        Files are not written to the project if a file of the same name exists.
-        'force' can be used to over-write files but files on the server that
-        are immutable cannot be over-written and will result in an error.
+        :param access_token: A valid DM API access token
+        :param project_id: The project where the files are to be written
+        :param project_files: A file or list of files. Leading paths are stripped
+            so the two file files ``['dir/file-a.txt', 'file-b.txt']`` would
+            be written to the same project directory, i.e. appearing as
+            ``/file-a.txt`` and ``/file-b.txt`` in the project
+        :param project_path: The path in the project to write the files.
+            The path is relative to the project root and must begin ``/``
+        :param force: Files are not written to the project if a file of the
+            same name exists. Here ``force`` can be used to over-write files.
+            Files on the server that are immutable cannot be over-written,
+            and doing so will result in an error
+        :param timeout_s: The underlying request timeout
         """
 
         assert access_token
@@ -386,7 +432,7 @@ class DmApi:
                and project_path.startswith('/')
 
         if not DmApi._dm_api_url:
-            return DmApiRv(success=False, msg={'msg': 'No API URL defined'})
+            return DmApiRv(success=False, msg={'error': 'No API URL defined'})
 
         # If we're not forcing the files collect the names
         # of every file on the path - we use this to skip files that
@@ -425,7 +471,7 @@ class DmApi:
             # whether we end up sending it or not.
             if not os.path.isfile(src_file):
                 return DmApiRv(success=False,
-                               msg={'msg': f'No such file ({src_file})'})
+                               msg={'error': f'No such file ({src_file})'})
             if os.path.basename(src_file) not in existing_path_files:
                 ret_val = DmApi.\
                     _put_unmanaged_project_file(access_token,
@@ -449,7 +495,17 @@ class DmApi:
                                        project_path: str = '/',
                                        timeout_s: int = 4)\
             -> DmApiRv:
-        """Deletes an unmanaged project file on a path.
+        """Deletes an unmanaged project file, or list of files, on a project path.
+
+        :param access_token: A valid DM API access token
+        :param project_id: The project where the files are present
+        :param project_files: A file or list of files. Leading paths are stripped
+            so the two file files ``['dir/file-a.txt', 'file-b.txt']`` would
+            be expected to be in the same project directory, i.e. appearing as
+            ``/file-a.txt`` and ``/file-b.txt`` in the project
+        :param project_path: The path in the project where the files are located.
+            The path is relative to the project root and must begin ``/``
+        :param timeout_s: The underlying request timeout
         """
         assert access_token
         assert project_id
@@ -489,7 +545,13 @@ class DmApi:
                            include_hidden: bool = False,
                            timeout_s: int = 8)\
             -> DmApiRv:
-        """Gets the list of project files on a path.
+        """Gets a list of project files on a path.
+
+        :param access_token: A valid DM API access token
+        :param project_id: The project where the files are present
+        :param project_path: The path in the project to search for files.
+            The path is relative to the project root and must begin ``/``
+        :param timeout_s: The underlying request timeout
         """
         assert access_token
         assert project_id
@@ -518,6 +580,14 @@ class DmApi:
             -> DmApiRv:
         """Get a single unmanaged file from a project path, saving it to
         the filename defined in local_file.
+
+        :param access_token: A valid DM API access token
+        :param project_id: The project where the files are present
+        :param project_file: The name of the file to get
+        :param local_file: The name to use to write the file to on the client
+        :param project_path: The path in the project to search for files.
+            The path is relative to the project root and must begin ``/``
+        :param timeout_s: The underlying request timeout
         """
         assert access_token
         assert project_id
@@ -553,10 +623,23 @@ class DmApi:
                                               project_path: str = '/',
                                               timeout_s: int = 8)\
             -> DmApiRv:
-        """Get a single unmanaged file from a project path using an Instance-generated
-        callback token. The file is saved to the filename defined in local_file.
-        This method does not require authentication but does require a token,
-        obtained when the job was executed and a callback given and token requested.
+        """Like :py:meth:`~DmApi.get_unmanaged_project_file()`, this method
+        get a single unmanaged file from a project path. The method uses an
+        Instance-generated callback token rathgr than a user-access token.
+
+        This method is particularly useful in callback routines where a user
+        access token may not be available. Callback tokens expire and can be
+        deleted and so this function should only be used when a user access
+        token is not available.
+
+        :param token: A DM-generated token, optionally generated when
+            launching instances in the project
+        :param project_id: The project where the files are present
+        :param project_file: The name of the file to get
+        :param local_file: The name to use to write the file to on the client
+        :param project_path: The path in the project to search for files.
+            The path is relative to the project root and must begin ``/``
+        :param timeout_s: The underlying request timeout
         """
         assert token
         assert project_id
@@ -588,15 +671,32 @@ class DmApi:
                            access_token: str,
                            project_id: str,
                            name: str,
+                           specification: Dict[str, Any],
                            callback_url: Optional[str] = None,
                            callback_context: Optional[str] = None,
                            generate_callback_token: bool = False,
-                           specification: Optional[Dict[str, Any]] = None,
                            debug: Optional[str] = None,
                            timeout_s: int = 4)\
             -> DmApiRv:
-        """Instantiates a Job (based on the latest Job application ID
-        and version known to the API).
+        """Instantiates a Job Instance in a Project.
+
+        :param access_token: A valid DM API access token
+        :param project_id: The project where the files are present
+        :param name: A name to associate with the Job
+        :param specification: The Job specification, it must contain
+            keys that define the Job's ``collection``, ``job name`` and
+            ``version``. Job-specific variables are passed in using a ``variables``
+            map in the specification
+        :param callback_url: An optional URL capable of handling Job callbacks.
+            Must be set if ``generate_callback_token`` is used
+        :param callback_context: An optional context string passed to the
+            callback URL
+        :param generate_callback_token: True to instruct the DM to generate
+            a token that can be used with some methods instead of a
+            user access token
+        :param debug: Used to prevent the automatic removal of the Job instance.
+            Only use this if you need to
+        :param timeout_s: The underlying request timeout
         """
 
         assert access_token
@@ -612,10 +712,10 @@ class DmApi:
             # Failed calling the server.
             # Incorrect URL, bad token or server out of action?
             return DmApiRv(success=False,
-                           msg={'msg': 'Failed getting Job operator version'})
+                           msg={'error': 'Failed getting Job operator version'})
         if not job_application_version:
             return DmApiRv(success=False,
-                           msg={'msg': 'No Job operator installed'})
+                           msg={'error': 'No Job operator installed'})
 
         data: Dict[str, Any] =\
             {'application_id': _DM_JOB_APPLICATION_ID,
@@ -642,6 +742,9 @@ class DmApi:
     def get_available_projects(cls, access_token: str, timeout_s: int = 4)\
             -> DmApiRv:
         """Gets information about all projects available to you.
+
+        :param access_token: A valid DM API access token
+        :param timeout_s: The underlying request timeout
         """
         assert access_token
 
@@ -658,6 +761,10 @@ class DmApi:
                     timeout_s: int = 4)\
             -> DmApiRv:
         """Gets detailed information about a specific project.
+
+        :param access_token: A valid DM API access token
+        :param project_id: The specific project to retrieve
+        :param timeout_s: The underlying request timeout
         """
         assert access_token
         assert project_id
@@ -674,7 +781,11 @@ class DmApi:
                      instance_id: str,
                      timeout_s: int = 4)\
             -> DmApiRv:
-        """Gets information about an Application/Job instance.
+        """Gets information about an instance (Application or Job).
+
+        :param access_token: A valid DM API access token
+        :param instance_id: The specific instance to retrieve
+        :param timeout_s: The underlying request timeout
         """
         assert access_token
         assert instance_id
@@ -692,6 +803,9 @@ class DmApi:
                               timeout_s: int = 4)\
             -> DmApiRv:
         """Gets information about all instances available to you.
+
+        :param access_token: A valid DM API access token
+        :param timeout_s: The underlying request timeout
         """
         assert access_token
         assert project_id
@@ -710,7 +824,16 @@ class DmApi:
                         instance_id: str,
                         timeout_s: int = 4)\
             -> DmApiRv:
-        """Deletes an Application/Job instance.
+        """Deletes an Instance (Application or Job).
+
+        When instances are deleted the container is removed along with
+        the instance-specific directory that is automatically created
+        in the root of the project. Any files in the instance-specific
+        directory wil be removed.
+
+        :param access_token: A valid DM API access token
+        :param instance_id: The instance to delete
+        :param timeout_s: The underlying request timeout
         """
         assert access_token
         assert instance_id
@@ -727,7 +850,14 @@ class DmApi:
                               token: str,
                               timeout_s: int = 4)\
             -> DmApiRv:
-        """Deletes an Application/Job instance.
+        """Deletes a DM API Instance **callback token**. This API method is not
+        authenticated and therefore does not need an access token. Once the token is
+        deleted no further calls to :py:meth:`DmApi.get_unmanaged_project_file_with_token()`
+        will be possible. Once deleted the token cannot be re-instantiated.
+
+        :param instance_id: A valid DM API instance
+        :param token: The callback Token associated with the instance
+        :param timeout_s: The API request timeout
         """
         assert instance_id
         assert token
@@ -767,7 +897,10 @@ class DmApi:
     @synchronized
     def get_available_jobs(cls, access_token: str, timeout_s: int = 4)\
             -> DmApiRv:
-        """Gets a summary list of avaailble Jobs.
+        """Gets a summary list of available Jobs.
+
+        :param access_token: A valid DM API access token.
+        :param timeout_s: The API request timeout
         """
         assert access_token
 
@@ -782,6 +915,10 @@ class DmApi:
             -> DmApiRv:
         """Gets detailed information about a specific Job
         using the numeric Job record identity
+
+        :param access_token: A valid DM API access token.
+        :param job_id: The numeric Job identity
+        :param timeout_s: The API request timeout
         """
         assert access_token
         assert job_id > 0
@@ -801,7 +938,13 @@ class DmApi:
                         timeout_s: int = 4)\
             -> DmApiRv:
         """Gets detailed information about a specific Job
-        using the collection, name and version
+        using the ``collection``, ``name`` and ``version``
+
+        :param access_token: A valid DM API access token.
+        :param job_collection: The Job collection, e.g. ``im-test``
+        :param job_name: The Job name, e.g. ``nop``
+        :param job_version: The Job version, e.g. ``1.0.0``
+        :param timeout_s: The API request timeout
         """
         assert access_token
         assert job_collection
@@ -825,8 +968,14 @@ class DmApi:
                         impersonate: Optional[str] = None,
                         timeout_s: int = 4)\
             -> DmApiRv:
-        """Adds or removes the 'become-admin' state of your account,
-        assuming you can do this.
+        """Adds or removes the ``become-admin`` state of your account.
+        Only users whose accounts offer administrative capabilities
+        can use this method.
+
+        :param access_token: A valid DM API access token.
+        :param admin: True to set admin state
+        :param imperrsonate: An optional username to switch to
+        :param timeout_s: The API request timeout
         """
         assert access_token
 
